/*
	GNUJSP - a free JSP implementation
	Copyright (C) 1998-1999, Vincent Partington <vinny@klomp.org>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package	org.gjt.jsp;

import	java.io.File;
import	java.io.IOException;
import	java.io.PrintWriter;
import	java.lang.reflect.Field;
import	java.lang.reflect.InvocationTargetException;
import	java.lang.reflect.Method;
import	java.util.Enumeration;
import	java.util.Hashtable;
import	java.util.StringTokenizer;
import	java.util.Vector;
import	javax.servlet.Servlet;
import	javax.servlet.ServletConfig;
import	javax.servlet.ServletContext;
import	javax.servlet.ServletException;
import	javax.servlet.http.HttpServlet;
import	javax.servlet.http.HttpServletRequest;
import	javax.servlet.http.HttpServletResponse;
import	java.io.InputStream;
import	java.io.OutputStream;
import	java.io.FileInputStream;
import	java.io.FileNotFoundException;
import	java.net.URLEncoder;

public class JSPServlet extends HttpServlet {
	boolean			correctlyInitialized = false;
	File			repository;
	String[]		compiler;
	boolean			checkDependencies = true;
	boolean			debug = false;
	String			defaultContentType = "text/html";
	boolean			cache_simple_html = false;
	boolean			checkUri = false;
	String			jspExtension = ".jsp";
	String			pageBase = "";

	JSPClassLoader	jspClassLoader;
	Hashtable		jspServlets;

	public void init(ServletConfig config) throws ServletException {
		super.init(config);

		String				s, s2;
		Vector				v;
		int					i;
		StringTokenizer		toker;

		// get repository where generated java and class files are stored
		s = config.getInitParameter("repository");
		if(s == null) {
			throw new ServletException("mandatory init parameter repository has not been specified");
		}
		repository = new File(s);
		try {
			repository = new File(repository.getCanonicalPath());
		} catch(IOException ioexc) {
			repository = new File(repository.getAbsolutePath());
		}

		// get compiler command line
		s = config.getInitParameter("compiler");
		if(s == null) {
			s = "builtin-javac -classpath %classpath%" + File.pathSeparator +
				"%repository% -d %repository% -deprecation %source%";
		}
		v = new Vector();
		toker = new StringTokenizer(s);
		while(toker.hasMoreTokens()) {
			v.addElement(toker.nextToken());
		}
		if(v.size() <= 1) {
			throw new ServletException("init parameter compiler does not specify the compiler's parameters");
		}
		compiler = new String[v.size()];
		v.copyInto(compiler);		

		// get check dependency settings
		s = config.getInitParameter("checkdependencies");
		if(s != null && (s.equalsIgnoreCase("false") || s.equalsIgnoreCase("no"))) {
			checkDependencies = false;
		}

		// get debug setting
		s = config.getInitParameter("debug");
		if(s != null && (s.equalsIgnoreCase("true") || s.equalsIgnoreCase("yes"))) {
			debug = true;
		}

		// get default content-type
		s = config.getInitParameter("defaultcontenttype");
		if(s != null) {
			defaultContentType = s;
		}

		// get cache_simple_html setting
		s = config.getInitParameter("cache_simple_html");
		if(s != null && (s.equalsIgnoreCase("true") || s.equalsIgnoreCase("yes"))) {
			cache_simple_html = true;
		}

		// get checkuri, extension and pagebase parameters
		// FIXME: these are hacks that will disappear when the
		//        Servlet API provides a secure and portable way
		//        to get the JSP source.
		s = config.getInitParameter("checkuri");
		if(s != null && (s.equalsIgnoreCase("true") || s.equalsIgnoreCase("yes"))) {
			checkUri = true;
		}

		s = config.getInitParameter("extension");
		if(s != null) {
			if(s.charAt(0) == '.') {
				jspExtension = s;
			} else {
				jspExtension = "." + s;
			}
		}

 		s = config.getInitParameter("pagebase");
		if(s != null) {
			pageBase = s;
		}

		jspClassLoader = new JSPClassLoader(repository, config, debug);
		jspServlets = new Hashtable();
		correctlyInitialized = true;
	}


    protected void copyInputStreamToOutputStream( OutputStream out, InputStream in )
    {
		byte    buf[] = new byte[ 4096 ];
		int     iRead;

		try
		{
			while( (iRead = in.read(buf,0,4096)) > 0 )
			{
				out.write( buf, 0, iRead );
			}
		}
		catch( IOException e )
		{
			// ignore
		}
    }


	protected void service(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
		int				i, j;
		File			jspFile;
		String			s, className;
		Servlet			srv;
		JSPCompiler		jspCompiler;

		if(!correctlyInitialized) {
			throw new ServletException("This servlet has not been initialized correctly");
		}

		// used by cache_simple_html to track if cache file should be
		// returned or discarded.
		String	strHtmlCacheFile = null;
		boolean	bDiscardHtmlCache = false;

		try {
			// check URI for security
			if(checkUri) {
				s = urlDecode(request.getRequestURI());
				i = s.indexOf("/servlet/");
				j = s.indexOf("/gnujsp/");
				if(i >= 0 && j >= 0 && i < j) {
					throw new JSPException("Direct requests to /servlet/gnujsp are not allowed.");
				}
			}

			// determine jsp file and class name
			jspFile = null;
			s = request.getServletPath();
			if(s == null) {
				s = request.getRequestURI();
			}
			if(s != null) {
				if(s.endsWith(jspExtension)) {
					s = getServletContext().getRealPath(s);
					if(s != null) {
						jspFile = new File(s);
					}
				} else {
					s = null;
				}
			}
			if(s == null) {
				s = request.getPathTranslated();
				if(s == null) {
					throw new JSPException(HttpServletResponse.SC_NOT_FOUND, "No JSP page specified");
				}
				if(!s.endsWith(jspExtension)) {
					throw new JSPException(HttpServletResponse.SC_FORBIDDEN, "The requested file does not have the required extension " + jspExtension);
				}
				// FIXME: should urlDecode() be used here? Seems to be necessary for ServletExec, but will it work with % in the filename?
				jspFile = new File(pageBase + urlDecode(s));
			}
			try {
				jspFile = new File(jspFile.getCanonicalPath());
			} catch(IOException ioexc) {
				jspFile = new File(jspFile.getAbsolutePath());
			}

			className = JSPCompiler.getClassNameForJspFile(jspFile);

			// try to load jsp servlet from cache or class file
			srv = (Servlet) jspServlets.get(className);
			if(srv == null) {
				srv = loadServlet(className);
				// checkDependencies has already been called
				if( srv == null )
				{
					// could not load servlet, so cache file must be old
					bDiscardHtmlCache = true;
				}
			}
			else
			{
				// unload servlet if dependencies demand it
				if(checkDependencies && srv != null) {
					if(checkServletDependencies(srv.getClass())) {
						unloadServlet(srv);
						refreshClassLoader(srv.getClass());
						srv = null;

						// servlet is outdated, so cache file must be also.
						bDiscardHtmlCache = true;
					}
				}
			}

			// compile & load jsp servlet if we haven't loaded it yet
			if(srv == null) {
				jspCompiler = new JSPCompiler(this, jspFile, className, request);
				jspCompiler.compile();

				srv = loadServlet(className);
				if(srv == null) {
					bDiscardHtmlCache = true;

					throw new JSPException("Could not load jsp servlet class " + className);
				}
			}
		} catch(JSPException jspexc) {
			if(debug) {
				getServletConfig().getServletContext().log(jspexc.getHttpErrorCode() + " JSP compile-time error: " + jspexc.getMessage());
			}
			response.setStatus(jspexc.getHttpErrorCode(), "JSP compile-time error");
			response.setContentType("text/html");
			PrintWriter out = response.getWriter();
			out.println("<HTML><HEAD><TITLE>" + jspexc.getHttpErrorCode() + " JSP compile-time error</TITLE></HEAD><BODY>" +
						"<H2>" + jspexc.getHttpErrorCode() + " JSP compile-time error</H2>" +
						"The JSP page you requested could not be served because the following error(s) occured:<BR><PRE>");
			out.println(jspexc.getMessage());
			out.println("</PRE></BODY></HTML>");
			out.close();
			return;
		}

		// should we try to return the cached html file first?
		Class	classSrv = srv.getClass();
		int		eMethod = Utils.getMethodEnum( request.getMethod() );
		if( cache_simple_html &&
			((eMethod & (Utils.METHOD_GET | Utils.METHOD_HEAD)) > 0) )
		{
			// generate filename, taking querystring and
			// character encoding into consideration.
			// ie. snoop.jsp_xxxx_en.html
			StringBuffer	strAppend = new StringBuffer();
			if( request.getQueryString() != null )
			{
				strAppend.append( "." );
				strAppend.append( URLEncoder.encode(request.getQueryString()) );
			}
			if( request.getCharacterEncoding() != null )
			{
				strAppend.append( "." );
				strAppend.append( URLEncoder.encode(response.getCharacterEncoding()) );
			}
			strAppend.append( ".html" );
			strHtmlCacheFile = jspFile.toString() + strAppend.toString();
				
			// try to find cache file
			try
			{
				boolean	bSendCache = false;
				File	fileCache = new File( strHtmlCacheFile );
				long lRecycle = 0;

				// determine what we should do
				if( fileCache.exists() == true )
				{
					if( fileCache.canRead() == false )
					{
						getServletConfig().getServletContext().log( "Can't read cache file " + strHtmlCacheFile );
					}
					else
					{
						// check if cache file is older than class file
						if( fileCache.lastModified() < jspFile.lastModified() )
						{
							bDiscardHtmlCache = true;
						}
						else
						{
							// check if the cache file needs to be recycled
							lRecycle = getServletCacheRecycleMillis( classSrv );
							if( lRecycle > 0 )
							{
								lRecycle += fileCache.lastModified();
							}
							if( (lRecycle > 0) && (lRecycle < System.currentTimeMillis()) )
							{
								bDiscardHtmlCache = true;
							}
							else
							{
								bSendCache = true;
							}
						}
					}
				}
				else
				{
					// no file, no need to discard
					bDiscardHtmlCache = false;
				}

				// should we discard?
				if( bDiscardHtmlCache == true )
				{
					// yes, delete all files matching our cache pattern
					String	strCachePath = jspFile.getPath();
					strCachePath = strCachePath.substring( 0, strCachePath.lastIndexOf(File.separator) );
					//getServletConfig().getServletContext().log( "Trying to delete cached files from " + strCachePath );

					Dirlist dirlist = new Dirlist( strCachePath );
					String	filenames[] = dirlist.list( jspFile.getName() + "*.html" );

					for( i=0; i<filenames.length; i++ )
					{
						File	file = new File( strCachePath, filenames[i] );

						if( file.delete() == false )
						{
							getServletConfig().getServletContext().log( "Failed to delete html cache file " + file.toString() );
						}
					}
				}

				// should we send?
				if( bSendCache )
				{
					// try to send cache file contents
					FileInputStream	readCache = new FileInputStream( fileCache );
					// set appropriate headers
					response.setContentType( getServletContentType(classSrv) );
					response.setContentLength( (int)fileCache.length() );
					response.setDateHeader( "Last-Modified", fileCache.lastModified() );
					if( lRecycle > 0 )
					{
						response.setDateHeader( "Expires", lRecycle );
					}

					// finally, try to send output
					OutputStream out = response.getOutputStream();
					if( eMethod == Utils.METHOD_GET )
					{
						// check for conditional GET (If-Modified-Since)
						if( Utils.IsCacheFileNewerThanRequest(fileCache.lastModified(),request.getDateHeader("If-Modified-Since")) == false )
						{
							// send 304
							response.sendError( HttpServletResponse.SC_NOT_MODIFIED );
						}	
						else
						{
							// send output
							copyInputStreamToOutputStream( out, readCache );
						}
					}

					// close up locks
					readCache.close();
					out.close();

					return;
				}
			}
			catch( FileNotFoundException e )
			{
				// file not found? ignore
			}
		}	// if( cache_simple_html )

		// run jsp servlet
		// but, should we cache the servlet's output?
		if( cache_simple_html &&
			getServletCacheSimpleHtml(classSrv) &&
			((eMethod & (Utils.METHOD_GET | Utils.METHOD_HEAD)) > 0) )
		{
			// capture output for cache
			CacheResponse	cacheResponse = new CacheResponse( response );
			
			srv.service( request, cacheResponse );

			// should we write to file?
			if( cacheResponse.getCantCache() == false )
			{
				cacheResponse.saveCachedResponse( strHtmlCacheFile,
												  getServletCacheRecycleMillis(classSrv) );
			}

			// finally, send response
			if( eMethod == Utils.METHOD_GET )
			{
				cacheResponse.sendCachedResponse();
			}
		}
		else
		{
			// no need to capture output, just return response
			srv.service(request, response);
		}
	}

	public void destroy() {
		super.destroy();

		if(correctlyInitialized) {
			Enumeration		e;
			Servlet			srv;
	
			e = jspServlets.elements();
			while(e.hasMoreElements()) {
				srv = (Servlet) e.nextElement();
				unloadServlet(srv);
			}
		}
	}

	public String getServletInfo() {
		return "GNUJSP JSPServlet";
	}


	protected boolean getServletCacheSimpleHtml( Class c )
	{
		Field		fld;

		try
		{
			fld = c.getField( "__cache_simple_html" );
			return fld.getBoolean( null );
		}
		catch(IllegalAccessException fiaexc)
		{
			if( debug )
			{
				getServletConfig().getServletContext().log( "Jsp servlet class " + c.getName() + " does not have a public static __cache_simple_html field" );
			}
		}
		catch(NoSuchFieldException nsfexc)
		{
			if( debug )
			{
				getServletConfig().getServletContext().log( "Jsp servlet class " + c.getName() + " does not have a __cache_simple_html field of type boolean" );
			}
		}
		catch(SecurityException secexc)
		{
			if( debug )
			{
				getServletConfig().getServletContext().log( "Jsp servlet class " + c.getName() + " has a security problem with its __cache_simple_html field" );
			}
		}

		return false;
	}


	protected long getServletCacheRecycleMillis( Class c )
	{
		Field		fld;

		try
		{
			fld = c.getField( "__cache_recycle_millis" );
			return fld.getLong( null );
		}
		catch(IllegalAccessException fiaexc)
		{
			if( debug )
			{
				getServletConfig().getServletContext().log( "Jsp servlet class " + c.getName() + " does not have a public static __cache_recycle_millis field" );
			}
		}
		catch(NoSuchFieldException nsfexc)
		{
			if( debug )
			{
				getServletConfig().getServletContext().log( "Jsp servlet class " + c.getName() + " does not have a __cache_recycle_millis field of type long" );
			}
		}
		catch(SecurityException secexc)
		{
			if( debug )
			{
				getServletConfig().getServletContext().log( "Jsp servlet class " + c.getName() + " has a security problem with its __cache_recycle_millis field" );
			}
		}

		return 0;
	}


	protected String getServletContentType( Class c )
	{
		Field		fld;

		try
		{
			fld = c.getField( "__content_typeDirective" );
			return (String)fld.get( null );
		}
		catch(IllegalAccessException fiaexc)
		{
			if( debug )
			{
				getServletConfig().getServletContext().log( "Jsp servlet class " + c.getName() + " does not have a public static __content_typeDirective field" );
			}
		}
		catch(NoSuchFieldException nsfexc)
		{
			if( debug )
			{
				getServletConfig().getServletContext().log( "Jsp servlet class " + c.getName() + " does not have a __content_typeDirective field of type String" );
			}
		}
		catch(SecurityException secexc)
		{
			if( debug )
			{
				getServletConfig().getServletContext().log( "Jsp servlet class " + c.getName() + " has a security problem with its __content_typeDirective field" );
			}
		}

		return "";
	}


	private Servlet loadServlet(String className) throws JSPException {
		Class		c;
		Field		fld;
		Servlet		srv;

		// load servlet class
		try {
			c = jspClassLoader.loadClass(className);
			if(c == null)
				return null;
		} catch(ClassNotFoundException cnfexc) {
			return null;
		}

		// check compiler version dependency
		try {
			fld = c.getField("__compilerVersionNr");
			if(fld.getInt(null) != JSPCompiler.COMPILER_VERSION_NR) {
				refreshClassLoader(c);
				return null;
			}
		} catch(IllegalAccessException fiaexc) {
			throw new JSPException("Jsp servlet class " + className + " does not have a public static __compilerVersionNr field");
		} catch(NoSuchFieldException nsfexc) {
			throw new JSPException("Jsp servlet class " + className + " does not have a __compilerVersionNr field of type int");
		} catch(SecurityException secexc) {
			throw new JSPException("Jsp servlet class " + className + " has a security problem with its __compilerVersionNr field");
		}

		// check servlet's other dependencies
		if(checkDependencies && checkServletDependencies(c)) {
			refreshClassLoader(c);
			return null;
		}

		// instantiate servlet
		try {
			srv = (Servlet) c.newInstance();
		} catch(IllegalAccessException iaexc) {
			throw new JSPException("Could not instantiate jsp servlet class " + className + " because of an illegal access exception");
		} catch(InstantiationException iexc) {
			throw new JSPException("Could not instantiate jsp servlet class " + className + " because it is an interface or an abstract class");
		}

		// init servlet
		try {
			srv.init(getServletConfig());
		} catch(Exception exc) {}

		jspServlets.put(className, srv);

		return srv;
	}

	private void unloadServlet(Servlet srv) {
		try {
			srv.destroy();
		} catch(Exception exc) { }

		jspServlets.remove(srv.getClass().getName());
	}

	private boolean checkServletDependencies(Class c) throws JSPException {
		Method			meth;

		try {
			// invoke dependency checking method
			meth = c.getMethod("__checkDependencies", new Class[0]);
			return ((Boolean) meth.invoke(null, new Object[0])).booleanValue();
		} catch(IllegalAccessException iaexc) {
			throw new JSPException("Jsp servlet class " + c.getName() + " does not have a public static __checkDependencies() method");
		} catch(InvocationTargetException itexc) {
			throw new JSPException("Jsp servlet class " + c.getName() + "'s __checkDependencies() method threw an exception: " + itexc.getTargetException());
		} catch(NoSuchMethodException nsfexc) {
			throw new JSPException("Jsp servlet class " + c.getName() + " does not have a __checkDependencies() method that returns a boolean");
		} catch(SecurityException secexc) {
			throw new JSPException("Jsp servlet class " + c.getName() + " has a security problem with its __checkDependencies() method");
		}
	}

	private void refreshClassLoader(Class c) {
		if(c.getClassLoader() == jspClassLoader) {
			jspClassLoader = new JSPClassLoader(repository, getServletConfig(), debug);
		}
	}

	private static String urlDecode(String val) {
        StringBuffer	buf = new StringBuffer(val.length());
        char			c;

		for (int i = 0; i < val.length(); i++) {
			c = val.charAt(i);
			if(c == '%') {
				try {
					buf.append((char)Integer.parseInt(val.substring(i+1,i+3),16));
					i += 2;
					continue;
				} catch(Exception e) { }
			} else if(c == '+') {
				buf.append(' ');
				continue;
			}
 			buf.append(c);
		}
        return buf.toString();
	}
}
